@@ -103,51 +103,6 @@ def main_count(db_path, table, *, where=None, distinct=None):
         print(db.count(table, where=where, distinct=distinct))
 
 
-@cli.command(name="create")
-@click.argument("db_path")
-@click.argument("table")
-@click.argument("payload", nargs=-1)
-@click.option(
-    "--auto-timestamp",
-    is_flag=True,
-    default=True,
-    help=(
-        "Automatically populate `created_at` and `last_updated_at` columns with "
-        + "current time."
-    ),
-)
-def main_create_wrapper(*args, **kwargs):
-    """
-    Create a new row non-interactively.
-
-    PAYLOAD should be a list of space-separated key-value pairs, e.g.
-
-        isqlite create --db db.sqlite3 my_table a=1 b=2
-    """
-    return main_create(*args, **kwargs)
-
-
-def main_create(db_path, table, payload, *, auto_timestamp=True):
-    if not payload:
-        report_error_and_exit("payload must not be empty")
-
-    payload_as_map = {}
-    for key_value in payload:
-        key, value = key_value.split("=")
-        payload_as_map[key] = value
-
-    if auto_timestamp:
-        auto_timestamp_columns = ["created_at", "last_updated_at"]
-    else:
-        auto_timestamp_columns = []
-
-    with Database(db_path) as db:
-        pk = db.create(
-            table, payload_as_map, auto_timestamp_columns=auto_timestamp_columns
-        )
-        print(f"Row {pk} created.")
-
-
 @cli.command(name="create-table")
 @click.argument("db_path")
 @click.argument("table")
@@ -216,9 +171,9 @@ def main_delete(db_path, table, pk=None, *, where="", no_confirm=False):
             if not no_confirm:
                 n = db.count(table, where=where)
                 if not where:
-                    msg = f"Are you sure you wish to delete ALL {n} row(s) from {table!r}?"
+                    msg = f"Are you sure you wish to delete ALL {pluralize(n, 'row')} from {table!r}?"
                 else:
-                    msg = f"Are you sure you wish to delete {n} row(s) from {table!r}?"
+                    msg = f"Are you sure you wish to delete {pluralize(n, 'row')} from {table!r}?"
 
                 if not click.confirm(msg):
                     print()
@@ -232,7 +187,67 @@ def main_delete(db_path, table, pk=None, *, where="", no_confirm=False):
 
             db.delete(table, where=where)
             print()
-            print(f"{n} row(s) from table {table!r} deleted.")
+            print(f"{pluralize(n, 'row')} from table {table!r} deleted.")
+
+
+@cli.command(name="diff")
+@click.argument("db_path")
+@click.argument("schema_path")
+@click.option("--table", default=None)
+def main_diff_wrapper(*args, **kwargs):
+    """
+    Diff the database against the Python schema.
+    """
+    main_diff(*args, **kwargs)
+
+
+def main_diff(db_path, schema_path, table):
+    schema = get_schema_from_path(schema_path)
+    with Database(db_path, readonly=True) as db:
+        _diff(db, schema, table)
+
+
+def _diff(db, schema, table):
+    diff = db.diff(schema, table=table)
+    if not diff:
+        print("Nothing to migrate: database matches schema.")
+        return None
+
+    tables_created = 0
+    tables_dropped = 0
+    columns_dropped = 0
+    printed = False
+    grouped_diff = group_diff_by_table(diff)
+    for table, table_diff in sorted(grouped_diff.items(), key=lambda kv: kv[0]):
+        if printed:
+            print()
+        else:
+            printed = True
+
+        print(f"Table {blue(table)}")
+        for op in table_diff:
+            if isinstance(op, migrations.DropColumnMigration):
+                columns_dropped += 1
+            elif isinstance(op, migrations.DropTableMigration):
+                tables_dropped += 1
+            elif isinstance(op, migrations.CreateTableMigration):
+                tables_created += 1
+
+            print(f"- {op}")
+
+    print()
+    print()
+    print("Summary")
+    print(f"- {blue(pluralize(len(grouped_diff), 'table'))} affected.")
+    print(f"- {blue(pluralize(len(diff), 'operation'))} in total.")
+    if tables_created > 0:
+        print(f"- {blue(pluralize(tables_created, 'table'))} created.")
+    if tables_dropped > 0:
+        print(f"- {red(pluralize(tables_dropped, 'table'))} dropped.")
+    if columns_dropped > 0:
+        print(f"- {red(pluralize(columns_dropped, 'column'))} dropped.")
+
+    return (diff, tables_dropped or columns_dropped)
 
 
 @cli.command(name="drop-column")
@@ -257,7 +272,9 @@ def main_drop_column(db_path, table, column, *, no_confirm=False):
         with Database(db_path) as db:
             count = db.count(table)
 
-            print(f"WARNING: Table {table!r} contains {count} row(s) of data.")
+            print(
+                f"WARNING: Table {table!r} contains {pluralize(count, 'row')} of data."
+            )
             print()
             if not click.confirm("Are you sure you wish to drop this column?"):
                 print()
@@ -290,7 +307,9 @@ def main_drop_table(db_path, table, *, no_confirm=False):
     with Database(db_path) as db:
         if not no_confirm:
             count = db.count(table)
-            print(f"WARNING: Table {table!r} contains {count} row(s) of data.")
+            print(
+                f"WARNING: Table {table!r} contains {pluralize(count, 'row')} of data."
+            )
             print()
             if not click.confirm("Are you sure you wish to drop this table?"):
                 print()
@@ -330,141 +349,49 @@ def main_get(db_path, table, pk, *, plain_foreign_keys=False):
             prettyprint_row(row)
 
 
-@cli.command(name="list")
+@cli.command(name="insert")
 @click.argument("db_path")
 @click.argument("table")
-@click.option("-w", "--where", default="")
-@click.option("-s", "--search")
-@click.option("--columns", multiple=True, default=[], help=HELP_COLUMNS)
-@click.option("--hide", multiple=True, default=[], help=HELP_HIDE)
-@click.option("-p", "--page", default=1, help=HELP_PAGE)
-@click.option("--limit", default=None, help=HELP_LIMIT)
-@click.option("--offset", default=None, help=HELP_OFFSET)
-@click.option("--order-by", multiple=True, default=[], help=HELP_ORDER_BY)
-@click.option("--desc", is_flag=True, default=False, help=HELP_DESC)
+@click.argument("payload", nargs=-1)
 @click.option(
-    "--plain-foreign-keys", is_flag=True, default=False, help=HELP_PLAIN_FOREIGN_KEYS
+    "--auto-timestamp",
+    is_flag=True,
+    default=True,
+    help=(
+        "Automatically populate `created_at` and `last_updated_at` columns with "
+        + "current time."
+    ),
 )
-def main_list_wrapper(*args, **kwargs):
+def main_insert_wrapper(*args, **kwargs):
     """
-    List the rows in the table, optionally filtered by a SQL clause.
+    Create a new row non-interactively.
+
+    PAYLOAD should be a list of space-separated key-value pairs, e.g.
+
+        isqlite insert --db db.sqlite3 my_table a=1 b=2
     """
-    main_list(*args, **kwargs)
+    return main_insert(*args, **kwargs)
 
 
-def main_list(
-    db_path,
-    table,
-    *,
-    where=None,
-    search="",
-    columns=[],
-    hide=[],
-    page=1,
-    limit=None,
-    offset=None,
-    order_by=[],
-    desc=False,
-    plain_foreign_keys=False,
-):
-    base_list(
-        db_path,
-        table,
-        where=where,
-        search=search,
-        columns=columns,
-        hide=hide,
-        page=page,
-        limit=limit,
-        offset=offset,
-        order_by=order_by,
-        desc=desc,
-        plain_foreign_keys=plain_foreign_keys,
-    )
+def main_insert(db_path, table, payload, *, auto_timestamp=True):
+    if not payload:
+        report_error_and_exit("payload must not be empty")
 
+    payload_as_map = {}
+    for key_value in payload:
+        key, value = key_value.split("=")
+        payload_as_map[key] = value
 
-def base_list(
-    db_path,
-    table,
-    *,
-    where,
-    search,
-    columns,
-    hide,
-    page,
-    limit,
-    offset,
-    order_by,
-    desc,
-    plain_foreign_keys,
-):
-    """
-    Base implementation shared by `main_list` and `main_search`
-    """
-    with Database(db_path, readonly=True) as db:
-        try:
-            rows = db.list(
-                table,
-                where=where,
-                order_by=order_by,
-                limit=limit,
-                offset=offset,
-                descending=desc if order_by else None,
-                get_related=not plain_foreign_keys,
-            )
-        except sqlite3.OperationalError:
-            # Because `get_related` uses SQL joins, it may cause 'ambiguous column'
-            # errors if the user-supplied WHERE clause has unqualified column names. So
-            # we simply retry on error with `get_related=False`.
-            rows = db.list(
-                table,
-                where=where,
-                order_by=order_by,
-                limit=limit,
-                offset=offset,
-                descending=desc if order_by else None,
-                get_related=False,
-            )
+    if auto_timestamp:
+        auto_timestamp_columns = ["created_at", "last_updated_at"]
+    else:
+        auto_timestamp_columns = []
 
-        if not plain_foreign_keys:
-            for row in rows:
-                for key, value in row.items():
-                    if isinstance(value, collections.OrderedDict):
-                        row[key] = get_column_as_string(value)
-
-        if search:
-            search = search.lower()
-            filtered_rows = []
-            for row in rows:
-                yes = False
-                for value in row.values():
-                    if isinstance(value, str) and search in value.lower():
-                        yes = True
-                    elif isinstance(value, collections.OrderedDict):
-                        for subvalue in value.values():
-                            if isinstance(subvalue, str) and search in subvalue.lower():
-                                yes = True
-                                break
-
-                    if yes:
-                        break
-
-                if yes:
-                    filtered_rows.append(row)
-            rows = filtered_rows
-
-        if not rows:
-            if search:
-                print(
-                    f"No rows found in table {table!r} "
-                    + f"matching search query {search!r}."
-                )
-            elif where:
-                print(f"No rows found in table {table!r} with constraint {where!r}.")
-            else:
-                print(f"No row founds in table {table!r}.")
-        else:
-            prettyprint_rows(rows, columns=columns, hide=hide, page=page)
+    with Database(db_path) as db:
+        pk = db.insert(
+            table, payload_as_map, auto_timestamp_columns=auto_timestamp_columns
+        )
+        print(f"Row {pk} created.")
 
 
 @cli.command(name="schema")
@@ -503,7 +430,7 @@ def main_schema(db_path, table=None, *, as_python=False):
             if as_python:
                 raise NotImplementedError
             else:
-                rows = db.list(
+                rows = db.select(
                     "sqlite_master",
                     where="type = 'table' AND name NOT LIKE 'sqlite_%'",
                     order_by="name",
@@ -545,71 +472,27 @@ def main_migrate_wrapper(*args, **kwargs):
 def main_migrate(db_path, schema_path, table, *, no_confirm, no_backup, debug):
     schema = get_schema_from_path(schema_path)
     with Database(db_path, transaction=False, debug=debug) as db:
-        diff = db.diff(schema, table=table)
-        if not diff:
-            print("Nothing to migrate: database matches schema.")
+        diff, data_dropped = _diff(db, schema, table)
+        if diff is None:
             return
 
+        if not no_confirm:
+            prompt = "Do you wish to perform the migration?"
+            if data_dropped:
+                prompt += red(" Some data may be lost.")
+
+            print()
+            if not click.confirm(prompt):
+                print()
+                print("Migration aborted.")
+                sys.exit(2)
+
         if not no_backup:
             _, backup_name = tempfile.mkstemp(
                 prefix="isqlite-backup-", suffix=".sqlite3"
             )
             shutil.copy2(db_path, backup_name)
 
-        tables_created = 0
-        tables_dropped = 0
-        printed = False
-        for table, table_diff in diff.items():
-            if printed:
-                print()
-            else:
-                printed = True
-
-            if len(table_diff) == 1 and isinstance(
-                table_diff[0], migrations.CreateTableMigration
-            ):
-                op = table_diff[0]
-                print(f"Create table {op.table_name}")
-                tables_created += 1
-            elif len(table_diff) == 1 and isinstance(
-                table_diff[0], migrations.DropTableMigration
-            ):
-                op = table_diff[0]
-                print(f"Drop table {op.table_name}")
-                tables_dropped += 1
-            else:
-                print(f"Table {table}")
-                for op in table_diff:
-                    print(f"- {op}")
-
-        print()
-        print()
-        if tables_created > 0:
-            print(f"Will create {tables_created} table(s).", end="")
-
-        if tables_dropped > 0:
-            print(f"Will drop {tables_dropped} table(s).", end="")
-
-        if diff:
-            total_ops = (
-                sum(len(table_diff) for table_diff in diff.values())
-                - tables_created
-                - tables_dropped
-            )
-            total_tables = len(diff) - tables_created - tables_dropped
-
-            if total_ops > 0:
-                print(
-                    f"Will perform {total_ops} operation(s) on {total_tables} table(s)."
-                )
-
-        if not no_confirm:
-            print()
-            if not click.confirm("Do you wish to perform the migration? "):
-                print()
-                print("Migration aborted.")
-                sys.exit(2)
-
         try:
             db.apply_diff(diff)
         except Exception:
@@ -697,7 +580,7 @@ def main_reorder_columns(db_path, table, columns):
 )
 def main_search_wrapper(*args, **kwargs):
     """
-    Shorthand for `list <table> -s <query>`
+    Shorthand for `select <table> -s <query>`
     """
     main_search(*args, **kwargs)
 
@@ -717,7 +600,7 @@ def main_search(
     desc=False,
     plain_foreign_keys=False,
 ):
-    base_list(
+    base_select(
         db_path,
         table,
         where=where,
@@ -733,6 +616,143 @@ def main_search(
     )
 
 
+@cli.command(name="select")
+@click.argument("db_path")
+@click.argument("table")
+@click.option("-w", "--where", default="")
+@click.option("-s", "--search")
+@click.option("--columns", multiple=True, default=[], help=HELP_COLUMNS)
+@click.option("--hide", multiple=True, default=[], help=HELP_HIDE)
+@click.option("-p", "--page", default=1, help=HELP_PAGE)
+@click.option("--limit", default=None, help=HELP_LIMIT)
+@click.option("--offset", default=None, help=HELP_OFFSET)
+@click.option("--order-by", multiple=True, default=[], help=HELP_ORDER_BY)
+@click.option("--desc", is_flag=True, default=False, help=HELP_DESC)
+@click.option(
+    "--plain-foreign-keys", is_flag=True, default=False, help=HELP_PLAIN_FOREIGN_KEYS
+)
+def main_select_wrapper(*args, **kwargs):
+    """
+    List the rows in the table, optionally filtered by a SQL clause.
+    """
+    main_select(*args, **kwargs)
+
+
+def main_select(
+    db_path,
+    table,
+    *,
+    where=None,
+    search="",
+    columns=[],
+    hide=[],
+    page=1,
+    limit=None,
+    offset=None,
+    order_by=[],
+    desc=False,
+    plain_foreign_keys=False,
+):
+    base_select(
+        db_path,
+        table,
+        where=where,
+        search=search,
+        columns=columns,
+        hide=hide,
+        page=page,
+        limit=limit,
+        offset=offset,
+        order_by=order_by,
+        desc=desc,
+        plain_foreign_keys=plain_foreign_keys,
+    )
+
+
+def base_select(
+    db_path,
+    table,
+    *,
+    where,
+    search,
+    columns,
+    hide,
+    page,
+    limit,
+    offset,
+    order_by,
+    desc,
+    plain_foreign_keys,
+):
+    """
+    Base implementation shared by `main_search` and `main_select`.
+    """
+    with Database(db_path, readonly=True) as db:
+        try:
+            rows = db.select(
+                table,
+                where=where,
+                order_by=order_by,
+                limit=limit,
+                offset=offset,
+                descending=desc if order_by else None,
+                get_related=not plain_foreign_keys,
+            )
+        except sqlite3.OperationalError:
+            # Because `get_related` uses SQL joins, it may cause 'ambiguous column'
+            # errors if the user-supplied WHERE clause has unqualified column names. So
+            # we simply retry on error with `get_related=False`.
+            rows = db.select(
+                table,
+                where=where,
+                order_by=order_by,
+                limit=limit,
+                offset=offset,
+                descending=desc if order_by else None,
+                get_related=False,
+            )
+
+        if not plain_foreign_keys:
+            for row in rows:
+                for key, value in row.items():
+                    if isinstance(value, collections.OrderedDict):
+                        row[key] = get_column_as_string(value)
+
+        if search:
+            search = search.lower()
+            filtered_rows = []
+            for row in rows:
+                yes = False
+                for value in row.values():
+                    if isinstance(value, str) and search in value.lower():
+                        yes = True
+                    elif isinstance(value, collections.OrderedDict):
+                        for subvalue in value.values():
+                            if isinstance(subvalue, str) and search in subvalue.lower():
+                                yes = True
+                                break
+
+                    if yes:
+                        break
+
+                if yes:
+                    filtered_rows.append(row)
+            rows = filtered_rows
+
+        if not rows:
+            if search:
+                print(
+                    f"No rows found in table {table!r} "
+                    + f"matching search query {search!r}."
+                )
+            elif where:
+                print(f"No rows found in table {table!r} with constraint {where!r}.")
+            else:
+                print(f"No row founds in table {table!r}.")
+        else:
+            prettyprint_rows(rows, columns=columns, hide=hide, page=page)
+
+
 @cli.command(name="sql")
 @click.argument("db_path")
 @click.argument("query")
@@ -826,7 +846,7 @@ def prettyprint_rows(rows, *, columns=[], hide=[], page=1):
 
     if table_rows:
         print()
-        print(f"{len(table_rows)} row(s).")
+        print(f"{pluralize(len(table_rows), 'row')}.")
 
     if overflow or page > 1:
         print()
@@ -895,6 +915,28 @@ def get_schema_from_path(schema_path):
     return schema_module.SCHEMA
 
 
+def group_diff_by_table(diff):
+    diff_map = collections.defaultdict(list)
+    for op in diff:
+        diff_map[op.table_name].append(op)
+    return diff_map
+
+
+def red(s):
+    return click.style(s, fg="red")
+
+
+def blue(s):
+    return click.style(s, fg="blue")
+
+
+def pluralize(n, word):
+    if n == 1:
+        return f"{n} {word}"
+    else:
+        return f"{n} {word}s"
+
+
 def report_error_and_exit(message):
     print(f"Error: {message}", file=sys.stderr)
     sys.exit(1)
