@@ -2,14 +2,13 @@ import collections
 import sqlite3
 import textwrap
 import warnings
-from typing import Any, Dict, List, Optional, Set, Tuple, Union
+from typing import Any, Dict, List, Optional, Tuple, Union
 
-import attr
 import sqliteparser
 from sqliteparser import quote
 
 from . import migrations
-from .schema import Schema
+from .schema import Diff, Schema, diff_schemas, diff_tables, rename_column
 
 CURRENT_TIMESTAMP_SQL = "STRFTIME('%Y-%m-%d %H:%M:%f', 'now')"
 AUTO_TIMESTAMP_DEFAULT = ("created_at", "last_updated_at")
@@ -19,7 +18,6 @@ AUTO_TIMESTAMP_UPDATE_DEFAULT = ("last_updated_at",)
 # Type aliases
 Row = Dict[str, Any]
 Rows = List[Dict]
-Diff = Dict[str, List[migrations.MigrateOperation]]
 
 
 class Database:
@@ -31,10 +29,10 @@ class Database:
             ...
 
     On creation, the ``Database`` connection will open a SQL transaction which will be
-    either committed or rolled back by the context manager, depending on whether an
-    exception occurs.
+    either committed or rolled back at the end of the ``with`` statement, depending on
+    whether an exception occurs.
 
-    You can also use multiple transactions over the life of the connection::
+    You can also have multiple transactions over the life of the connection::
 
         with Database("db.sqlite3", transaction=False) as db:
             with db.transaction():
@@ -44,6 +42,13 @@ class Database:
                 ...
     """
 
+    connection: sqlite3.Connection
+    cursor: sqlite3.Cursor
+    debugger: Optional["Debugger"]
+    schema: Schema
+    insert_auto_timestamp_columns: List[str]
+    update_auto_timestamp_columns: List[str]
+
     def __init__(
         self,
         path: str,
@@ -53,7 +58,7 @@ class Database:
         readonly: Optional[bool] = None,
         uri: bool = False,
         cached_statements: int = 100,
-        create_auto_timestamp_columns: List[str] = [],
+        insert_auto_timestamp_columns: List[str] = [],
         update_auto_timestamp_columns: List[str] = [],
     ) -> None:
         """
@@ -86,8 +91,8 @@ class Database:
         :param uri: If true, the ``path`` argument is interpreted as a URI rather than a
             file path.
         :param cached_statements: Passed on to ``sqlite3.connect``.
-        :param create_auto_timestamp_columns: A default value for
-            ``auto_timestamp_columns`` in ``create`` and ``create_many``. Usually set to
+        :param insert_auto_timestamp_columns: A default value for
+            ``auto_timestamp_columns`` in ``insert`` and ``insert_many``. Usually set to
             ``["created_at", "last_updated_at"]`` in conjunction with a schema defined
             using ``AutoTable``.
         :param update_auto_timestamp_columns: A default value for
@@ -116,7 +121,7 @@ class Database:
             else:
                 path = f"file:{path}"
 
-        self.create_auto_timestamp_columns = create_auto_timestamp_columns
+        self.insert_auto_timestamp_columns = insert_auto_timestamp_columns
         self.update_auto_timestamp_columns = update_auto_timestamp_columns
 
         self.connection = sqlite3.connect(
@@ -142,7 +147,7 @@ class Database:
 
         self.refresh_schema()
 
-    def list(
+    def select(
         self,
         table: str,
         *,
@@ -188,7 +193,7 @@ class Database:
         else:
             if descending is not None:
                 raise ISqliteApiError(
-                    "The `descending` parameter to `list` requires the `order_by` "
+                    "The `descending` parameter to `select` requires the `order_by` "
                     + "parameter to be set."
                 )
             order_clause = ""
@@ -201,7 +206,7 @@ class Database:
         else:
             if offset is not None:
                 raise ISqliteApiError(
-                    "The `offset` parameter to `list` requires the `limit` parameter "
+                    "The `offset` parameter to `select` requires the `limit` parameter "
                     + "to be set."
                 )
 
@@ -231,38 +236,39 @@ class Database:
         *,
         where: str = "",
         values: Dict[str, Any] = {},
+        order_by: Optional[str] = None,
+        descending: Optional[bool] = None,
         get_related: Union[List[str], bool] = [],
-    ) -> Row:
+    ) -> Optional[Row]:
         """
-        Retrieve a single row from the database table and return it as an ``OrderedDict``
-        object.
+        Retrieve a single row from the database table and return it as an
+        ``OrderedDict`` object.
+
+        Equivalent to ``Database.select(*args, **kwargs)[0]`` except that ``None`` is
+        returned if no matching row is found, and the SQLite engine only fetches a
+        single row from the database.
 
         :param table: The database table to query. WARNING: This value is directly
             interpolated into the SQL statement. Do not pass untrusted input, to avoid
             SQL injection attacks.
-        :param where: Same as for ``Database.list``.
-        :param values: Same as for ``Database.list``.
-        :param get_related: Same as for ``Database.list``.
+        :param where: Same as for ``Database.select``.
+        :param values: Same as for ``Database.select``.
+        :param order_by: Same as for ``Database.select``.
+        :param descending: Same as for ``Database.select``.
+        :param get_related: Same as for ``Database.select``.
         """
-        where_clause = f"WHERE {where}" if where else ""
+        rows = self.select(
+            table,
+            where=where,
+            values=values,
+            order_by=order_by,
+            descending=descending,
+            limit=1,
+            get_related=get_related,
+        )
+        return rows[0] if rows else None
 
-        if get_related:
-            columns, joins = self._get_related_columns_and_joins(table, get_related)
-            row = self.sql(
-                f"SELECT {columns} FROM {quote(table)} {joins} {where_clause}",
-                values,
-                multiple=False,
-            )
-        else:
-            row = self.sql(
-                f"SELECT * FROM {quote(table)} {where_clause}",
-                values,
-                multiple=False,
-            )
-
-        return row
-
-    def get_by_pk(self, table: str, pk: int, **kwargs) -> Row:
+    def get_by_pk(self, table: str, pk: int, **kwargs) -> Optional[Row]:
         """
         Retrieve a single row from the database table by its primary key.
 
@@ -275,30 +281,35 @@ class Database:
         pk_column = f"{quote(table)}.rowid"
         return self.get(table, where=f"{pk_column} = :pk", values={"pk": pk}, **kwargs)
 
-    def get_or_create(self, table: str, data: Row, **kwargs) -> Row:
+    def get_or_insert(self, table: str, data: Row, **kwargs) -> Row:
         """
         Retrieve a single row from the database table matching the parameters in
-        ``data``. If no such row exists, create it and return it.
+        ``data``. If no such row exists, insert it and return it.
+
+        Not to be confused with ``insert_and_get``, which unconditionally inserts a row
+        and returns it.
 
         :param table: The database table to query. WARNING: This value is directly
             interpolated into the SQL statement. Do not pass untrusted input, to avoid
             SQL injection attacks.
         :param data: The parameters to match the database row. All required columns of
-            the table must be included, or else the internal call to ``Database.create``
+            the table must be included, or else the internal call to ``Database.insert``
             will fail.
-        :param kwargs: Additional arguments to pass on to ``Database.create``. If the
+        :param kwargs: Additional arguments to pass on to ``Database.insert``. If the
             database row already exists, these arguments are ignored.
         """
         if not data:
             raise ISqliteError(
-                "The `data` parameter to `get_or_create` cannot be empty."
+                "The `data` parameter to `get_or_insert` cannot be empty."
             )
 
         query = " AND ".join(f"{key} = :{key}" for key in data)
         row = self.get(table, where=query, values=data)
         if row is None:
-            pk = self.create(table, data, **kwargs)
-            return self.get_by_pk(table, pk)
+            pk = self.insert(table, data, **kwargs)
+            row = self.get_by_pk(table, pk)
+            assert row is not None
+            return row
         else:
             return row
 
@@ -316,8 +327,8 @@ class Database:
         :param table: The database table to query. WARNING: This value is directly
             interpolated into the SQL statement. Do not pass untrusted input, to avoid
             SQL injection attacks.
-        :param where: Same as for ``Database.list``.
-        :param values: Same as for ``Database.list``.
+        :param where: Same as for ``Database.select``.
+        :param values: Same as for ``Database.select``.
         :param distinct: Only count rows with distinct values of this column.
         """
         where_clause = f"WHERE {where}" if where else ""
@@ -330,15 +341,17 @@ class Database:
         )
         return result[0]
 
-    def create(
+    def insert(
         self,
         table: str,
         data: Row,
         *,
-        auto_timestamp_columns: Union[List[str], bool] = False,
+        auto_timestamp_columns: Union[List[str], bool] = True,
     ) -> int:
         """
-        Insert a new row.
+        Insert a new row and return its primary key.
+
+        To get the contents of the row after it is inserted, use ``insert_and_get``.
 
         :param table: The database table. WARNING: This value is directly interpolated
             into the SQL statement. Do not pass untrusted input, to avoid SQL injection
@@ -347,11 +360,11 @@ class Database:
             values.
         :param auto_timestamp_columns: A list of columns into which to insert the
             current date and time, as an ISO 8601 timestamp. If true, it defaults to
-            the value of ``create_auto_timestamp_columns`` passed to ``__init__``.
+            the value of ``insert_auto_timestamp_columns`` passed to ``__init__``.
         """
         if isinstance(auto_timestamp_columns, bool):
             if auto_timestamp_columns is True:
-                auto_timestamp_columns_list = self.create_auto_timestamp_columns
+                auto_timestamp_columns_list = self.insert_auto_timestamp_columns
             else:
                 auto_timestamp_columns_list = []
         else:
@@ -380,12 +393,38 @@ class Database:
         self.cursor.execute(sql, values)
         return self.cursor.lastrowid
 
-    def create_many(
+    def insert_and_get(
+        self,
+        table: str,
+        data: Row,
+        *,
+        auto_timestamp_columns: Union[List[str], bool] = True,
+    ) -> Row:
+        """
+        Same as ``insert``, except it fetches the row after it is inserted and returns
+        it. Note that this requires an extra SQL query.
+
+        Not to be confused with ``get_or_insert``, which first tries to fetch a matching
+        row and only inserts a new row if no matching one exists.
+
+        The returned row may differ from ``data`` for two reasons:
+
+        - The ``auto_timestamp_columns`` parameter causes isqlite to insert values
+          into additional columns besides those in ``data``.
+        - SQLite will supply default values if possible for any columns of the table
+          omitted from ``data``.
+        """
+        pk = self.insert(table, data, auto_timestamp_columns=auto_timestamp_columns)
+        row = self.get_by_pk(table, pk)
+        assert row is not None
+        return row
+
+    def insert_many(
         self,
         table: str,
         data: Rows,
         *,
-        auto_timestamp_columns: Union[List[str], bool] = False,
+        auto_timestamp_columns: Union[List[str], bool] = True,
     ) -> None:
         """
         Insert multiple rows at once.
@@ -393,7 +432,7 @@ class Database:
         Equivalent to::
 
             for row in data:
-                db.create(table, row)
+                db.insert(table, row)
 
         but more efficient.
         """
@@ -402,7 +441,7 @@ class Database:
 
         if isinstance(auto_timestamp_columns, bool):
             if auto_timestamp_columns is True:
-                auto_timestamp_columns_list = self.create_auto_timestamp_columns
+                auto_timestamp_columns_list = self.insert_auto_timestamp_columns
             else:
                 auto_timestamp_columns_list = []
         else:
@@ -437,7 +476,7 @@ class Database:
         *,
         where: str = "",
         values: Dict[str, Any] = {},
-        auto_timestamp_columns: Union[List[str], bool] = False,
+        auto_timestamp_columns: Union[List[str], bool] = True,
     ) -> None:
         """
         Update an existing row.
@@ -447,15 +486,16 @@ class Database:
             attacks.
         :param data: The columns to update, as a dictionary from column names to column
             values.
-        :param where: Restrict the set of rows to update. Same as for ``Database.list``.
-        :param values: Same as for ``Database.list``.
-        :param auto_timestamp_columns: Same as for ``Database.create``, except that if
+        :param where: Restrict the set of rows to update. Same as for
+            ``Database.select``.
+        :param values: Same as for ``Database.select``.
+        :param auto_timestamp_columns: Same as for ``Database.insert``, except that if
             the same column appears in both ``values`` and ``auto_timestamp_columns``,
             the timestamp will be inserted instead of the value.
         """
         if isinstance(auto_timestamp_columns, bool):
             if auto_timestamp_columns is True:
-                auto_timestamp_columns_list = self.create_auto_timestamp_columns
+                auto_timestamp_columns_list = self.insert_auto_timestamp_columns
             else:
                 auto_timestamp_columns_list = []
         else:
@@ -513,10 +553,10 @@ class Database:
         :param table: The database table. WARNING: This value is directly interpolated
             into the SQL statement. Do not pass untrusted input, to avoid SQL injection
             attacks.
-        :param where: Same as for ``Database.list``, except that it is required, to
+        :param where: Same as for ``Database.select``, except that it is required, to
             avoid accidentally deleting every row in a table. If you indeed wish to
             delete every row, then pass ``where="1"``.
-        :param values: Same as for ``Database.list``.
+        :param values: Same as for ``Database.select``.
         """
         if not where:
             raise ISqliteApiError(
@@ -589,7 +629,7 @@ class Database:
         """
         if isinstance(columns, str):
             raise ISqliteApiError(
-                "second argument to DatabaseMigrator.create_table must be a list, "
+                "second argument to `Database.create_table` must be a list, "
                 + "not a string"
             )
 
@@ -741,31 +781,17 @@ class Database:
         :param schema: The Python schema to compare against the database.
         :param table: The table to diff. If empty, the entire database will be diffed.
         """
-        tables_in_db = self._get_sql_schema()
-        tables_in_schema = schema.tables if not table else [schema[table]]
+        self.refresh_schema()
+        if table:
+            try:
+                old_schema = self.schema[table]
+                new_schema = schema[table]
+            except KeyError:
+                raise TableDoesNotExistError(table)
 
-        diff: Diff = collections.defaultdict(list)
-        for table_in_schema in tables_in_schema:
-            name = table_in_schema.name
-            if name in tables_in_db:
-                columns_in_database = tables_in_db.pop(table_in_schema.name).columns
-                columns_in_schema = [
-                    column for column in table_in_schema.columns.values()
-                ]
-                self._diff_table(diff, name, columns_in_database, columns_in_schema)
-            else:
-                diff[table_in_schema.name].append(
-                    migrations.CreateTableMigration(
-                        table_in_schema.name,
-                        [str(column) for column in table_in_schema.columns.values()],
-                    )
-                )
-
-        if table is None:
-            for name in tables_in_db:
-                diff[name].append(migrations.DropTableMigration(name))
-
-        return diff
+            return diff_tables(old_schema, new_schema)
+        else:
+            return diff_schemas(self.schema, schema)
 
     def apply_diff(self, diff: Diff) -> None:
         """
@@ -786,32 +812,29 @@ class Database:
             )
 
         with self.transaction(disable_foreign_keys=True):
-            for table_diff in diff.values():
-                for op in table_diff:
-                    if isinstance(op, migrations.CreateTableMigration):
-                        self.create_table(op.table_name, op.columns)
-                    elif isinstance(op, migrations.DropTableMigration):
-                        self.drop_table(op.table_name)
-                    elif isinstance(op, migrations.AlterColumnMigration):
-                        self.alter_column(
-                            op.table_name,
-                            op.column.name,
-                            str(op.column.definition)
-                            if op.column.definition is not None
-                            else "",
-                        )
-                    elif isinstance(op, migrations.AddColumnMigration):
-                        self.add_column(op.table_name, op.column)
-                    elif isinstance(op, migrations.DropColumnMigration):
-                        self.drop_column(op.table_name, op.column_name)
-                    elif isinstance(op, migrations.ReorderColumnsMigration):
-                        self.reorder_columns(op.table_name, op.column_names)
-                    elif isinstance(op, migrations.RenameColumnMigration):
-                        self.rename_column(
-                            op.table_name, op.old_column_name, op.new_column_name
-                        )
-                    else:
-                        raise ISqliteError("unknown migration op type")
+            for op in diff:
+                if isinstance(op, migrations.CreateTableMigration):
+                    self.create_table(op.table_name, op.columns)
+                elif isinstance(op, migrations.DropTableMigration):
+                    self.drop_table(op.table_name)
+                elif isinstance(op, migrations.AlterColumnMigration):
+                    self.alter_column(
+                        op.table_name,
+                        op.column_name,
+                        op.column_definition,
+                    )
+                elif isinstance(op, migrations.AddColumnMigration):
+                    self.add_column(op.table_name, op.column)
+                elif isinstance(op, migrations.DropColumnMigration):
+                    self.drop_column(op.table_name, op.column_name)
+                elif isinstance(op, migrations.ReorderColumnsMigration):
+                    self.reorder_columns(op.table_name, op.column_names)
+                elif isinstance(op, migrations.RenameColumnMigration):
+                    self.rename_column(
+                        op.table_name, op.old_column_name, op.new_column_name
+                    )
+                else:
+                    raise ISqliteError("unknown migration op type")
 
             self.refresh_schema()
 
@@ -837,10 +860,10 @@ class Database:
         using ``Database.sql`` or in an external database connection, you may need to
         call this method for correct behavior.
 
-        The internal schema is used by the ``get_related`` functionality of ``list`` and
-        ``get``.
+        The internal schema is used by the ``get_related`` functionality of ``select``
+        and ``get``.
         """
-        self.schema = self._get_sql_schema()
+        self.schema = self._get_schema_from_database()
 
     def transaction(
         self, *, disable_foreign_keys: bool = False
@@ -925,67 +948,6 @@ class Database:
 
         self.close()
 
-    def _diff_table(
-        self,
-        diff: Diff,
-        table_name: str,
-        columns_in_database: List[sqliteparser.ast.Column],
-        columns_in_schema: List[sqliteparser.ast.Column],
-    ) -> Diff:
-        columns_in_database_map = {
-            column.name: i for i, column in enumerate(columns_in_database)
-        }
-        renamed_columns: Set[str] = set()
-        reordered = False
-        for new_index, column in enumerate(columns_in_schema):
-            old_index = columns_in_database_map.get(column.name)
-            if old_index is None:
-                if new_index < len(columns_in_database) and is_renamed_column(
-                    column, columns_in_database[new_index]
-                ):
-                    old_column_name = columns_in_database[new_index].name
-                    renamed_columns.add(old_column_name)
-                    diff[table_name].append(
-                        migrations.RenameColumnMigration(
-                            table_name, old_column_name, column.name
-                        )
-                    )
-                else:
-                    diff[table_name].append(
-                        migrations.AddColumnMigration(table_name, str(column))
-                    )
-                continue
-
-            if old_index != new_index:
-                reordered = True
-
-            old_column = columns_in_database[old_index]
-            if old_column != column:
-                diff[table_name].append(
-                    migrations.AlterColumnMigration(table_name, column)
-                )
-
-        columns_in_schema_map = {
-            column.name: i for i, column in enumerate(columns_in_schema)
-        }
-        for column in columns_in_database:
-            if (
-                column.name not in columns_in_schema_map
-                and column.name not in renamed_columns
-            ):
-                diff[table_name].append(
-                    migrations.DropColumnMigration(table_name, column.name)
-                )
-
-        if reordered:
-            diff[table_name].append(
-                migrations.ReorderColumnsMigration(
-                    table_name, [column.name for column in columns_in_schema]
-                )
-            )
-
-        return diff
-
     def _migrate_table(self, name: str, columns: List[str], *, select: str) -> None:
         # This procedure is copied from https://sqlite.org/lang_altertable.html
         # Create the new table under a temporary name.
@@ -1079,14 +1041,15 @@ class Database:
         )
         return columns, joins
 
-    def _get_sql_schema(self) -> Dict[str, sqliteparser.ast.Column]:
-        tables_in_db = {
-            row["name"]: sqliteparser.parse(row["sql"])[0]
-            for row in self.list(
-                "sqlite_master", where="type = 'table' AND NOT name LIKE 'sqlite_%'"
-            )
-        }
-        return tables_in_db
+    def _get_schema_from_database(self) -> Schema:
+        return Schema(
+            [
+                sqliteparser.parse(row["sql"])[0]
+                for row in self.select(
+                    "sqlite_master", where="type = 'table' AND NOT name LIKE 'sqlite_%'"
+                )
+            ]
+        )
 
 
 class TransactionContextManager:
@@ -1127,7 +1090,7 @@ def ordered_dict_row_factory(cursor: sqlite3.Cursor, row: Tuple[Any]) -> Row:
         name = column[0]
         value = row[i]
 
-        # When `get_related` is passed to `Database.get` or `Database.list`, the SQL
+        # When `get_related` is passed to `Database.get` or `Database.select`, the SQL
         # query fetches columns from foreign key relationships and names them with the
         # format {original_table_column}____{related_table_column}, i.e. if the
         # `students` table has a `major` column that points to the `departments` table,
@@ -1180,84 +1143,6 @@ def get_foreign_key_model(column: sqliteparser.ast.Column) -> Optional[str]:
     return None
 
 
-def is_renamed_column(
-    column_in_schema: sqliteparser.ast.Column,
-    column_in_database: sqliteparser.ast.Column,
-) -> bool:
-    return column_in_schema == rename_column(
-        column_in_database, column_in_database.name, column_in_schema.name
-    )
-
-
-def rename_column(
-    column: sqliteparser.ast.Column,
-    old_name: str,
-    new_name: str,
-) -> sqliteparser.ast.ColumnDefinition:
-    renamer = ColumnRenamer(old_name, new_name)
-    return renamer.rename(column)
-
-
-class ColumnRenamer:
-    """
-    A class implementing the visitor pattern which renames all instances of a column's
-    name in the column's definition.
-    """
-
-    def __init__(self, old_name, new_name):
-        self.old_name = old_name
-        self.new_name = new_name
-
-    def rename(self, node):
-        if node is None:
-            return None
-
-        return node.accept(self)
-
-    def visit_column(self, node):
-        return attr.evolve(
-            node,
-            name=self.new_name if node.name == self.old_name else node.name,
-            definition=self.rename(node.definition),
-        )
-
-    def visit_column_definition(self, node):
-        return attr.evolve(node, constraints=list(map(self.rename, node.constraints)))
-
-    def visit_check_constraint(self, node):
-        return attr.evolve(node, expr=self.rename(node.expr))
-
-    def visit_named_constraint(self, node):
-        return attr.evolve(node, constraint=self.rename(node.constraint))
-
-    def visit_foreign_key_constraint(self, node):
-        columns = [
-            self.new_name if column == self.old_name else column
-            for column in node.columns
-        ]
-        return attr.evolve(node, columns=columns)
-
-    def visit_generated_column_constraint(self, node):
-        return attr.evolve(node, expression=self.rename(node.expression))
-
-    def visit_infix(self, node):
-        return attr.evolve(
-            node, left=self.rename(node.left), right=self.rename(node.right)
-        )
-
-    def visit_expression_list(self, node):
-        return attr.evolve(node, values=list(map(self.rename, node.values)))
-
-    def visit_identifier(self, node):
-        if node.value == self.old_name:
-            return sqliteparser.ast.Identifier(self.new_name)
-        else:
-            return node
-
-    def visit_default(self, node):
-        return node
-
-
 class Debugger:
     def execute(self, sql: str, values: Any) -> None:
         self._execute("Execute", sql, values)
